@@ -11,40 +11,41 @@
 # FOR A PARTICULAR PURPOSE.
 #
 ##############################################################################
-"""A foundation for relstorage adapter tests"""
+"""A foundation for RelStorage tests"""
 
-import itertools
-import time
-from relstorage.relstorage import RelStorage
-from relstorage.tests import fakecache
-
-from ZODB.DB import DB
-from ZODB.utils import p64
-from ZODB.FileStorage import FileStorage
 from persistent import Persistent
 from persistent.mapping import PersistentMapping
+from relstorage.tests import fakecache
+from ZODB.DB import DB
+from ZODB.serialize import referencesf
+from ZODB.tests import BasicStorage
+from ZODB.tests import ConflictResolution
+from ZODB.tests import MTStorage
+from ZODB.tests import PackableStorage
+from ZODB.tests import PersistentStorage
+from ZODB.tests import ReadOnlyStorage
+from ZODB.tests import StorageTestBase
+from ZODB.tests import Synchronization
+from ZODB.tests.MinPO import MinPO
+from ZODB.tests.StorageTestBase import zodb_pickle
+from ZODB.tests.StorageTestBase import zodb_unpickle
+from ZODB.utils import p64
+import random
+import time
 import transaction
 
-from ZODB.tests import StorageTestBase, BasicStorage, \
-     TransactionalUndoStorage, PackableStorage, \
-     Synchronization, ConflictResolution, HistoryStorage, \
-     IteratorStorage, RevisionStorage, PersistentStorage, \
-     MTStorage, ReadOnlyStorage, RecoveryStorage
 
-from ZODB.tests.MinPO import MinPO
-from ZODB.tests.StorageTestBase import zodb_unpickle, zodb_pickle
-from ZODB.serialize import referencesf
-
-
-class BaseRelStorageTests(StorageTestBase.StorageTestBase):
+class RelStorageTestBase(StorageTestBase.StorageTestBase):
 
     def make_adapter(self):
         # abstract method
         raise NotImplementedError
 
     def open(self, **kwargs):
+        from relstorage.storage import RelStorage
         adapter = self.make_adapter()
-        self._storage = RelStorage(adapter, pack_gc=True, **kwargs)
+        self._storage = RelStorage(adapter, **kwargs)
+        self._storage._batcher_row_limit = 1
 
     def setUp(self):
         self.open(create=1)
@@ -56,26 +57,20 @@ class BaseRelStorageTests(StorageTestBase.StorageTestBase):
         self._storage.cleanup()
 
 
-class RelStorageTests(
-    BaseRelStorageTests,
+class GenericRelStorageTests(
+    RelStorageTestBase,
     BasicStorage.BasicStorage,
-    TransactionalUndoStorage.TransactionalUndoStorage,
-    RevisionStorage.RevisionStorage,
     PackableStorage.PackableStorage,
-    PackableStorage.PackableUndoStorage,
     Synchronization.SynchronizedStorage,
     ConflictResolution.ConflictResolvingStorage,
-    HistoryStorage.HistoryStorage,
-    IteratorStorage.IteratorStorage,
-    IteratorStorage.ExtendedIteratorStorage,
     PersistentStorage.PersistentStorage,
     MTStorage.MTStorage,
-    ReadOnlyStorage.ReadOnlyStorage
+    ReadOnlyStorage.ReadOnlyStorage,
     ):
 
     def checkDropAndPrepare(self):
-        self._storage._adapter.drop_all()
-        self._storage._adapter.prepare_schema()
+        self._storage._adapter.schema.drop_all()
+        self._storage._adapter.schema.prepare()
 
     def checkCrossConnectionInvalidation(self):
         # Verify connections see updated state at txn boundaries
@@ -212,6 +207,35 @@ class RelStorageTests(
         self.assertEqual(len(got), len(data))
         self.assertEqual(got, data)
 
+    def check16MObject(self):
+        # Store 16 * 1024 * 1024 bytes in an object, then retrieve it
+        data = 'a 16 byte string' * (1024 * 1024)
+        oid = self._storage.new_oid()
+        self._dostoreNP(oid, data=data)
+        got, serialno = self._storage.load(oid, '')
+        self.assertEqual(len(got), len(data))
+        self.assertEqual(got, data)
+
+    def check99X1900Objects(self):
+        # Store 99 objects each with 1900 bytes.  This is intended
+        # to exercise possible buffer overfilling that the batching
+        # code might cause.
+        import transaction
+        data = '0123456789012345678' * 100
+        t = transaction.Transaction()
+        self._storage.tpc_begin(t)
+        oids = []
+        for i in range(99):
+            oid = self._storage.new_oid()
+            self._storage.store(oid, '\0'*8, data, '', t)
+            oids.append(oid)
+        self._storage.tpc_vote(t)
+        self._storage.tpc_finish(t)
+        for oid in oids:
+            got, serialno = self._storage.load(oid, '')
+            self.assertEqual(len(got), len(data))
+            self.assertEqual(got, data)
+
     def checkPreventOIDOverlap(self):
         # Store an object with a particular OID, then verify that
         # OID is not reused.
@@ -222,42 +246,39 @@ class RelStorageTests(
         self.assert_(oid1 < oid2, 'old OID %r should be less than new OID %r'
             % (oid1, oid2))
 
-    def check16MObject(self):
-        # Store 16 * 1024 * 1024 bytes in an object, then retrieve it
-        data = 'a 16 byte string' * (1024 * 1024)
-        oid = self._storage.new_oid()
-        self._dostoreNP(oid, data=data)
-        got, serialno = self._storage.load(oid, '')
-        self.assertEqual(len(got), len(data))
-        self.assertEqual(got, data)
-
-    def checkLoadFromCache(self):
+    def checkUseCache(self):
         # Store an object, cache it, then retrieve it from the cache
         self._storage._options.cache_servers = 'x:1 y:2'
         self._storage._options.cache_module_name = fakecache.__name__
+        self._storage._options.cache_prefix = 'zzz'
 
+        fakecache.data.clear()
         db = DB(self._storage)
         try:
             c1 = db.open()
-            self.assertEqual(c1._storage._cache_client.servers, ['x:1', 'y:2'])
-            fakecache.data.clear()
+            self.assert_(c1._storage._cache.clients_global_first[0].servers,
+                ['x:1', 'y:2'])
             r1 = c1.root()
-            # the root tid and state should now be cached
-            self.assertEqual(len(fakecache.data), 2)
+            # The root state and checkpoints should now be cached.
+            # A commit count *might* be cached depending on the ZODB version.
+            self.assertTrue('zzz:checkpoints' in fakecache.data)
+            self.assertEqual(sorted(fakecache.data.keys())[-1][:10],
+                'zzz:state:')
             r1['alpha'] = PersistentMapping()
-            self.assertFalse('commit_count' in fakecache.data)
             transaction.commit()
-            self.assertTrue('commit_count' in fakecache.data)
-            self.assertEqual(len(fakecache.data), 3)
+            self.assertEqual(len(fakecache.data.keys()), 5)
+
             oid = r1['alpha']._p_oid
+            got, serial = c1._storage.load(oid, '')
+            # another state should now be cached
+            self.assertEqual(len(fakecache.data.keys()), 5)
+
+            # make a change
+            r1['beta'] = 0
+            transaction.commit()
+            self.assertEqual(len(fakecache.data.keys()), 6)
 
             got, serial = c1._storage.load(oid, '')
-            # another tid and state should now be cached
-            self.assertEqual(len(fakecache.data), 5)
-
-            # load the object via loadSerial()
-            got2 = c1._storage.loadSerial(oid, serial)
-            self.assertEqual(got, got2)
 
             # try to load an object that doesn't exist
             self.assertRaises(KeyError, c1._storage.load, 'bad.oid.', '')
@@ -306,276 +327,117 @@ class RelStorageTests(
             self.assert_(c2 is c1)
             r = c2.root()
             self.assertEqual(r['alpha'], 1)
-            r['beta'] = 2
+            r['beta'] = PersistentMapping()
+            c2.add(r['beta'])
             transaction.commit()
             c2.close()
         finally:
             db.close()
 
-    def checkPollInterval(self, using_cache=False):
+    def checkPollInterval(self, shared_cache=True):
         # Verify the poll_interval parameter causes RelStorage to
         # delay invalidation polling.
         self._storage._options.poll_interval = 3600
         db = DB(self._storage)
         try:
-            c1 = db.open()
+            tm1 = transaction.TransactionManager()
+            c1 = db.open(transaction_manager=tm1)
             r1 = c1.root()
             r1['alpha'] = 1
-            transaction.commit()
+            tm1.commit()
 
-            c2 = db.open()
+            tm2 = transaction.TransactionManager()
+            c2 = db.open(transaction_manager=tm2)
             r2 = c2.root()
             self.assertEqual(r2['alpha'], 1)
+            self.assertFalse(c2._storage.need_poll())
+            self.assertTrue(c2._storage._poll_at > 0)
 
             r1['alpha'] = 2
-            # commit c1 without triggering c2.afterCompletion().
-            storage = c1._storage
-            t = transaction.Transaction()
-            storage.tpc_begin(t)
-            c1.commit(t)
-            storage.tpc_vote(t)
-            storage.tpc_finish(t)
+            # commit c1 without committing c2.
+            tm1.commit()
 
-            # flush invalidations to c2, but the poll timer has not
-            # yet expired, so the change to r2 should not be seen yet.
-            self.assertTrue(c2._storage._poll_at > 0)
-            if using_cache:
+            if shared_cache:
                 # The cache reveals that a poll is needed even though
                 # the poll timeout has not expired.
                 self.assertTrue(c2._storage.need_poll())
-                c2._flush_invalidations()
+                tm2.commit()
                 r2 = c2.root()
                 self.assertEqual(r2['alpha'], 2)
                 self.assertFalse(c2._storage.need_poll())
             else:
+                # The poll timeout has not expired, so no poll should occur
+                # yet, even after a commit.
                 self.assertFalse(c2._storage.need_poll())
-                c2._flush_invalidations()
+                tm2.commit()
                 r2 = c2.root()
                 self.assertEqual(r2['alpha'], 1)
 
             # expire the poll timer and verify c2 sees the change
             c2._storage._poll_at -= 3601
-            c2._flush_invalidations()
+            tm2.commit()
             r2 = c2.root()
             self.assertEqual(r2['alpha'], 2)
 
-            transaction.abort()
             c2.close()
             c1.close()
 
         finally:
             db.close()
 
-    def checkPollIntervalWithCache(self):
-        self._storage._options.cache_servers = 'x:1'
-        self._storage._options.cache_module_name = fakecache.__name__
-        fakecache.data.clear()
-        self.checkPollInterval(using_cache=True)
-
-
-    def checkTransactionalUndoIterator(self):
-        # this test overrides the broken version in TransactionalUndoStorage.
-
-        s = self._storage
-
-        BATCHES = 4
-        OBJECTS = 4
-
-        orig = []
-        for i in range(BATCHES):
-            t = transaction.Transaction()
-            tid = p64(i + 1)
-            s.tpc_begin(t, tid)
-            txn_orig = []
-            for j in range(OBJECTS):
-                oid = s.new_oid()
-                obj = MinPO(i * OBJECTS + j)
-                revid = s.store(oid, None, zodb_pickle(obj), '', t)
-                txn_orig.append((tid, oid, revid))
-            serials = s.tpc_vote(t)
-            if not serials:
-                orig.extend(txn_orig)
-            else:
-                # The storage provided revision IDs after the vote
-                serials = dict(serials)
-                for tid, oid, revid in txn_orig:
-                    self.assertEqual(revid, None)
-                    orig.append((tid, oid, serials[oid]))
-            s.tpc_finish(t)
-
-        i = 0
-        for tid, oid, revid in orig:
-            self._dostore(oid, revid=revid, data=MinPO(revid),
-                          description="update %s" % i)
-
-        # Undo the OBJECTS transactions that modified objects created
-        # in the ith original transaction.
-
-        def undo(i):
-            info = s.undoInfo()
-            t = transaction.Transaction()
-            s.tpc_begin(t)
-            base = i * OBJECTS + i
-            for j in range(OBJECTS):
-                tid = info[base + j]['id']
-                s.undo(tid, t)
-            s.tpc_vote(t)
-            s.tpc_finish(t)
-
-        for i in range(BATCHES):
-            undo(i)
-
-        # There are now (2 + OBJECTS) * BATCHES transactions:
-        #     BATCHES original transactions, followed by
-        #     OBJECTS * BATCHES modifications, followed by
-        #     BATCHES undos
-
-        iter = s.iterator()
-        offset = 0
-
-        eq = self.assertEqual
-
-        for i in range(BATCHES):
-            txn = iter[offset]
-            offset += 1
-
-            tid = p64(i + 1)
-            eq(txn.tid, tid)
-
-            L1 = [(rec.oid, rec.tid, rec.data_txn) for rec in txn]
-            L2 = [(oid, revid, None) for _tid, oid, revid in orig
-                  if _tid == tid]
-
-            eq(L1, L2)
-
-        for i in range(BATCHES * OBJECTS):
-            txn = iter[offset]
-            offset += 1
-            eq(len([rec for rec in txn if rec.data_txn is None]), 1)
-
-        for i in range(BATCHES):
-            txn = iter[offset]
-            offset += 1
-
-            # The undos are performed in reverse order.
-            otid = p64(BATCHES - i)
-            L1 = [rec.oid for rec in txn]
-            L2 = [oid for _tid, oid, revid in orig if _tid == otid]
-            L1.sort()
-            L2.sort()
-            eq(L1, L2)
-
-        self.assertRaises(IndexError, iter.__getitem__, offset)
-
-    def checkNonASCIITransactionMetadata(self):
-        # Verify the database stores and retrieves non-ASCII text
-        # in transaction metadata.
-        ugly_string = ''.join(chr(c) for c in range(256))
+    def checkPollIntervalWithUnsharedCache(self):
+        self._storage._options.share_local_cache = False
+        self.checkPollInterval(shared_cache=False)
 
+    def checkCachePolling(self):
+        self._storage._options.poll_interval = 3600
+        self._storage._options.share_local_cache = False
         db = DB(self._storage)
         try:
-            c1 = db.open()
+            # Set up the database.
+            tm1 = transaction.TransactionManager()
+            c1 = db.open(transaction_manager=tm1)
             r1 = c1.root()
-            r1['alpha'] = 1
-            transaction.get().setUser(ugly_string)
-            transaction.commit()
-            r1['alpha'] = 2
-            transaction.get().note(ugly_string)
-            transaction.commit()
+            r1['obj'] = obj1 = PersistentMapping({'change': 0})
+            tm1.commit()
 
-            info = self._storage.undoInfo()
-            self.assertEqual(info[0]['description'], ugly_string)
-            self.assertEqual(info[1]['user_name'], '/ ' + ugly_string)
-        finally:
-            db.close()
+            # Load and change the object in an independent connection.
+            tm2 = transaction.TransactionManager()
+            c2 = db.open(transaction_manager=tm2)
+            r2 = c2.root()
+            r2['obj']['change'] = 1
+            tm2.commit()
+            # Now c2 has delta_after0.
+            self.assertEqual(len(c2._storage._cache.delta_after0), 1)
+            c2.close()
 
-    def checkPackGC(self, expect_object_deleted=True):
-        db = DB(self._storage)
-        try:
-            c1 = db.open()
-            r1 = c1.root()
-            r1['alpha'] = PersistentMapping()
-            transaction.commit()
+            # Change the object in the original connection.
+            c1.sync()
+            obj1['change'] = 2
+            tm1.commit()
 
-            oid = r1['alpha']._p_oid
-            r1['alpha'] = None
-            transaction.commit()
+            # Close the database connection to c2.
+            c2._storage._drop_load_connection()
 
-            # The object should still exist
-            self._storage.load(oid, '')
+            # Make the database connection to c2 reopen without polling.
+            c2._storage.load('\0' * 8, '')
+            self.assertTrue(c2._storage._load_transaction_open)
 
-            # Pack
-            now = packtime = time.time()
-            while packtime <= now:
-                packtime = time.time()
-            self._storage.pack(packtime, referencesf)
-            self._storage.sync()
+            # Open a connection, which should be the same connection
+            # as c2.
+            c3 = db.open(transaction_manager=tm2)
+            self.assertTrue(c3 is c2)
+            self.assertEqual(len(c2._storage._cache.delta_after0), 1)
 
-            if expect_object_deleted:
-                # The object should now be gone
-                self.assertRaises(KeyError, self._storage.load, oid, '')
-            else:
-                # The object should still exist
-                self._storage.load(oid, '')
-        finally:
-            db.close()
+            # Clear the caches (but not delta_after*)
+            c3._resetCache()
+            for client in c3._storage._cache.clients_local_first:
+                client.flush_all()
 
-    def checkPackGCDisabled(self):
-        self._storage._options.pack_gc = False
-        self.checkPackGC(expect_object_deleted=False)
+            obj3 = c3.root()['obj']
+            # Should have loaded the new object.
+            self.assertEqual(obj3['change'], 2)
 
-    def checkPackGCDryRun(self):
-        self._storage._options.pack_dry_run = True
-        self.checkPackGC(expect_object_deleted=False)
-
-    def checkPackOldUnreferenced(self):
-        db = DB(self._storage)
-        try:
-            c1 = db.open()
-            r1 = c1.root()
-            r1['A'] = PersistentMapping()
-            B = PersistentMapping()
-            r1['A']['B'] = B
-            transaction.get().note('add A then add B to A')
-            transaction.commit()
-
-            del r1['A']['B']
-            transaction.get().note('remove B from A')
-            transaction.commit()
-
-            r1['A']['C'] = ''
-            transaction.get().note('add C to A')
-            transaction.commit()
-
-            now = packtime = time.time()
-            while packtime <= now:
-                packtime = time.time()
-            self._storage.pack(packtime, referencesf)
-
-            # B should be gone, since nothing refers to it.
-            self.assertRaises(KeyError, self._storage.load, B._p_oid, '')
-
-        finally:
-            db.close()
-
-    def checkPackDutyCycle(self):
-        # Exercise the code in the pack algorithm that releases the
-        # commit lock for a time to allow concurrent transactions to commit.
-        self._storage._options.pack_batch_timeout = 0  # pause after every txn
-
-        slept = []
-        def sim_sleep(seconds):
-            slept.append(seconds)
-
-        db = DB(self._storage)
-        try:
-            # Pack
-            now = packtime = time.time()
-            while packtime <= now:
-                packtime = time.time()
-            self._storage.pack(packtime, referencesf, sleep=sim_sleep)
-
-            self.assertEquals(len(slept), 1)
         finally:
             db.close()
 
@@ -597,6 +459,91 @@ class RelStorageTests(
         finally:
             db.close()
 
+    def checkHistoryWithExtension(self):
+        # Verify the history method works with transactions that have
+        # extended info.
+        db = DB(self._storage)
+        try:
+            conn = db.open()
+            try:
+                conn.root()['pi'] = 3.14
+                transaction.get().setExtendedInfo("digits", 3)
+                transaction.commit()
+                history = self._storage.history(conn.root()._p_oid)
+                self.assertEqual(len(history), 1)
+                if self.keep_history:
+                    self.assertEqual(history[0]['digits'], 3)
+            finally:
+                conn.close()
+        finally:
+            db.close()
+
+    def checkPackDutyCycle(self):
+        # Exercise the code in the pack algorithm that releases the
+        # commit lock for a time to allow concurrent transactions to commit.
+        self._storage._options.pack_batch_timeout = 0  # pause after every txn
+
+        slept = []
+        def sim_sleep(seconds):
+            slept.append(seconds)
+
+        db = DB(self._storage)
+        try:
+            # add some data to be packed
+            c = db.open()
+            r = c.root()
+            r['alpha'] = PersistentMapping()
+            transaction.commit()
+            del r['alpha']
+            transaction.commit()
+
+            # Pack
+            now = packtime = time.time()
+            while packtime <= now:
+                packtime = time.time()
+            self._storage.pack(packtime, referencesf, sleep=sim_sleep)
+
+            self.assertTrue(len(slept) > 0)
+        finally:
+            db.close()
+
+    def checkPackKeepNewObjects(self):
+        # Packing should not remove objects created or modified after
+        # the pack time, even if they are unreferenced.
+        db = DB(self._storage)
+        try:
+            # add some data to be packed
+            c = db.open()
+            extra1 = PersistentMapping()
+            c.add(extra1)
+            extra2 = PersistentMapping()
+            c.add(extra2)
+            transaction.commit()
+
+            # Choose the pack time
+            now = packtime = time.time()
+            while packtime <= now:
+                time.sleep(0.1)
+                packtime = time.time()
+            while packtime == time.time():
+                time.sleep(0.1)
+
+            extra2.foo = 'bar'
+            extra3 = PersistentMapping()
+            c.add(extra3)
+            transaction.commit()
+
+            self._storage.pack(packtime, referencesf)
+
+            # extra1 should have been garbage collected
+            self.assertRaises(KeyError,
+                self._storage.load, extra1._p_oid, '')
+            # extra2 and extra3 should both still exist
+            self._storage.load(extra2._p_oid, '')
+            self._storage.load(extra3._p_oid, '')
+        finally:
+            db.close()
+
     def checkPackBrokenPickle(self):
         # Verify the pack stops with the right exception if it encounters
         # a broken pickle.
@@ -605,6 +552,107 @@ class RelStorageTests(
         self.assertRaises(UnpicklingError, self._storage.pack,
             time.time() + 10000, referencesf)
 
+    def checkBackwardTimeTravel(self):
+        # When a failover event causes the storage to switch to an
+        # asynchronous slave that is not fully up to date, the poller
+        # should notice that backward time travel has occurred and
+        # handle the situation by invalidating all objects that have
+        # changed in the interval. (Currently, we simply invalidate all
+        # objects when backward time travel occurs.)
+        import os
+        import shutil
+        import tempfile
+        from ZODB.FileStorage import FileStorage
+        db = DB(self._storage)
+        try:
+            c = db.open()
+            r = c.root()
+            r['alpha'] = PersistentMapping()
+            transaction.commit()
+
+            # To simulate failover to an out of date async slave, take
+            # a snapshot of the database at this point, change some
+            # object, then restore the database to its earlier state.
+
+            d = tempfile.mkdtemp()
+            try:
+                fs = FileStorage(os.path.join(d, 'Data.fs'))
+                fs.copyTransactionsFrom(c._storage)
+
+                r['beta'] = PersistentMapping()
+                transaction.commit()
+                self.assertTrue('beta' in r)
+
+                c._storage.zap_all()
+                c._storage.copyTransactionsFrom(fs)
+
+                fs.close()
+            finally:
+                shutil.rmtree(d)
+
+            # r should still be in the cache.
+            self.assertTrue('beta' in r)
+
+            # Now sync, which will call poll_invalidations().
+            c.sync()
+
+            # r should have been invalidated
+            self.assertEqual(r._p_changed, None)
+
+            # r should be reverted to its earlier state.
+            self.assertFalse('beta' in r)
+
+        finally:
+            db.close()
+
+    def checkBTreesLengthStress(self):
+        # BTrees.Length objects are unusual Persistent objects: they
+        # set _p_independent and they frequently invoke conflict
+        # resolution. Run a stress test on them.
+        updates_per_thread = 50
+        thread_count = 4
+
+        from BTrees.Length import Length
+        db = DB(self._storage)
+        try:
+            c = db.open()
+            try:
+                c.root()['length'] = Length()
+                transaction.commit()
+            finally:
+                c.close()
+
+            def updater():
+                thread_db = DB(self._storage)
+                for i in range(updates_per_thread):
+                    thread_c = thread_db.open()
+                    try:
+                        thread_c.root()['length'].change(1)
+                        time.sleep(random.random() * 0.05)
+                        transaction.commit()
+                    finally:
+                        thread_c.close()
+
+            import threading
+            threads = []
+            for i in range(thread_count):
+                t = threading.Thread(target=updater)
+                threads.append(t)
+            for t in threads:
+                t.start()
+            for t in threads:
+                t.join(120)
+
+            c = db.open()
+            try:
+                self.assertEqual(c.root()['length'](),
+                    updates_per_thread * thread_count)
+            finally:
+                transaction.abort()
+                c.close()
+
+        finally:
+            db.close()
 
 class DoubleCommitter(Persistent):
     """A crazy persistent class that changes self in __getstate__"""
@@ -612,89 +660,3 @@ class DoubleCommitter(Persistent):
         if not hasattr(self, 'new_attribute'):
             self.new_attribute = 1
         return Persistent.__getstate__(self)
-
-
-class IteratorDeepCompareUnordered:
-    # Like IteratorDeepCompare, but compensates for OID order
-    # differences in transactions.
-    def compare(self, storage1, storage2):
-        eq = self.assertEqual
-        iter1 = storage1.iterator()
-        iter2 = storage2.iterator()
-        for txn1, txn2 in itertools.izip(iter1, iter2):
-            eq(txn1.tid,         txn2.tid)
-            eq(txn1.status,      txn2.status)
-            eq(txn1.user,        txn2.user)
-            eq(txn1.description, txn2.description)
-
-            missing = object()
-            e1 = getattr(txn1, 'extension', missing)
-            if e1 is missing:
-                # old attribute name
-                e1 = txn1._extension
-            e2 = getattr(txn2, 'extension', missing)
-            if e2 is missing:
-                # old attribute name
-                e2 = txn2._extension
-            eq(e1, e2)
-
-            recs1 = [(r.oid, r) for r in txn1]
-            recs1.sort()
-            recs2 = [(r.oid, r) for r in txn2]
-            recs2.sort()
-            eq(len(recs1), len(recs2))
-            for (oid1, rec1), (oid2, rec2) in zip(recs1, recs2):
-                eq(rec1.oid,     rec2.oid)
-                eq(rec1.tid,  rec2.tid)
-                eq(rec1.version, rec2.version)
-                eq(rec1.data,    rec2.data)
-        # Make sure ther are no more records left in txn1 and txn2, meaning
-        # they were the same length
-        self.assertRaises(IndexError, iter1.next)
-        self.assertRaises(IndexError, iter2.next)
-        iter1.close()
-        iter2.close()
-
-
-
-class RecoveryStorageSubset(IteratorDeepCompareUnordered):
-    # The subset of RecoveryStorage tests that do not rely on version
-    # support.
-    pass
-
-for name, attr in RecoveryStorage.RecoveryStorage.__dict__.items():
-    if 'check' in name and 'Version' not in name:
-        setattr(RecoveryStorageSubset, name, attr)
-
-
-class ToFileStorage(BaseRelStorageTests, RecoveryStorageSubset):
-    def setUp(self):
-        self.open(create=1)
-        self._storage.zap_all()
-        self._dst = FileStorage("Dest.fs", create=True)
-
-    def tearDown(self):
-        self._storage.close()
-        self._dst.close()
-        self._storage.cleanup()
-        self._dst.cleanup()
-
-    def new_dest(self):
-        return FileStorage('Dest.fs')
-
-
-class FromFileStorage(BaseRelStorageTests, RecoveryStorageSubset):
-    def setUp(self):
-        self.open(create=1)
-        self._storage.zap_all()
-        self._dst = self._storage
-        self._storage = FileStorage("Source.fs", create=True)
-
-    def tearDown(self):
-        self._storage.close()
-        self._dst.close()
-        self._storage.cleanup()
-        self._dst.cleanup()
-
-    def new_dest(self):
-        return self._dst
